%% FastFourierExtension.m 
% Code to test the Fast Prolate Pseudoinverse and Fast Tikhonov 
% Regularization. This code generates the data for "The Fast Slepian 
% Transform" Figures 5, and 6. 
% After running this script, run FastFourierExtension_makefig56.m to  
% generate Figures 5, and 6. 
%
% Most recent change - 08/07/2017
%
% Copyright 2017, Santhosh Karnik
%
% This file is part of Fast Slepian Transform (FST) Toolbox version 1.0.
%
%    FST is free software: you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation, either version 3 of the License, or
%    (at your option) any later version.
%
%    FST is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License
%    along with FST.  If not, see <http://www.gnu.org/licenses/>.
%
% If you use this code in an academic paper, please cite our paper:
% S. Karnik, Z. Zhu, M. B. Wakin, J. K. Romberg, and M. A. Davenport., 
% The fast Slepian transform, Appl. Comput. Harmon. Anal. (2017), 
% http://dx.doi.org/10.1016/j.acha.2017.07.005

T = 1.5; % Extended interval half-length
cutoff = 1e-4; % Threshold for truncated pseudoinverse
alpha = 1e-8; % Regularization parameter for Tikhonov regularization
tolerance = 1e-5; % Tolerance for fast methods

% Function to reconstruct
% f(t) = sum a_k*exp(-|t-mu_k|/sigma_k) for random a_k, mu_k, sigma_k
rng(1);
K = 500;
mu = (2*rand(1,K)-1);
sigma = 0.1*rand(1,K)+0.01;
a = randn(1,K); a = a+0.5*sign(a);

% Sample f(t) at N equispaced points in [-1,1]
N = 2^25;
% fsamp = zeros(N,1);
fsamp = 5*(2*(0:N-1)'/N-1);
for k = 1:K,
    [1,k]
    fsamp = fsamp + a(k)*exp(-abs(((2*(0:N-1)'/N-1)-mu(k))./sigma(k)));
end

% Sample f(t)*chi_[-1,1](t) at N equispaced points in [-T,T]
tsampT = T*(2*(0:N-1)'/N-1);
indexT = (abs(tsampT) <= 1);
fsampT = zeros(N,1);
fsampT(indexT) = 5*tsampT(indexT);
for k = 1:K,
    [2,k]
    fsampT(indexT) = fsampT(indexT) + a(k)*exp(-abs((tsampT(indexT)-mu(k))./sigma(k)));
end

% Create arrays for statistics
vecM = [1:255,255,256:32:511,511,512:64:1023,1023,1024:128:2047,2047,2048:256:2560];
qmax = floor(log2(max(vecM)));
numtrials = 1;
Error_Fourier = zeros(1,length(vecM));
Error_FourierExtensionPinv = zeros(1,length(vecM));
Error_FourierExtensionTik = zeros(1,length(vecM));
Error_FourierExtensionPinvFast = zeros(1,length(vecM));
Error_FourierExtensionTikFast = zeros(1,length(vecM));
Time_Fourier = zeros(numtrials,length(vecM));
Time_FourierExtensionPinv = zeros(numtrials,length(vecM));
Time_FourierExtensionTik = zeros(numtrials,length(vecM));
Time_FourierExtensionPinvFast = zeros(numtrials,length(vecM));
Time_FourierExtensionTikFast = zeros(numtrials,length(vecM));
flag = false(1,length(vecM));

% Repeat coefficient calculation many times to get an average time
for ii = 1:numtrials,
    % For each value of M (Truncate series for f(t) to 2M+1 terms)
    for jj = 1:length(vecM),
        M = vecM(jj);
        [ii,M]
        
        q = floor(log2(M));
        Nsamp = N/2^(qmax-q);
        Dsamp = N/Nsamp;
        fdownsamp = fsamp(1:Dsamp:end);
        fdownsampT = fsampT(1:Dsamp:end);
        fftf = fft(fdownsamp);
        clear fftf
        % Compute Fourier Series coefficients
        tic
            % Approximate inner products using FFT
        fftf = fft(fdownsamp);
        fhat = (sqrt(2)/Nsamp)*((-1).^(-M:M)').*[fftf((Nsamp-M+1):Nsamp);fftf(1:(M+1))];
        Time_Fourier(ii,jj) = toc;
        clear fftf
        
        if(M < 4096)
            % Compute Fourier Extension coefficients using slow pseudoinverse
            tic
                % Approximate inner products using FFT
            fftfT = fft(fdownsampT);
            fhatT = (sqrt(2*T)/Nsamp)*((-1).^(-M:M)').*[fftfT((Nsamp-M+1):Nsamp);fftfT(1:(M+1))];
                % Solve system using truncated pseudoinverse
            if(M > 320)
                Koverestimate = min(2*M+1,ceil((2*M+1)/T + log(2*(2*M+1))*log(2./cutoff)/pi^2+3));
                [S,lambda] = myDPSS(2*M+1,1/(2*T),Koverestimate);
                Kactual = nnz(lambda > cutoff);
                ghatTpinv = S(:,1:Kactual)*((S(:,1:Kactual)'*fhatT)./lambda(1:Kactual));
            else
                B = toeplitz(sinc((0:(2*M))/T)/T);
                ghatTpinv = pinv(B,cutoff)*fhatT;
            end
            Time_FourierExtensionPinv(ii,jj) = toc;
            if(~isempty(lastwarn))
                flag(jj) = true;
                lastwarn('');
            end
            clear fftfT fhatT S lambda Koverestimate Kactual

            % Compute Fourier Extension coefficients using slow Tikhonov regularization
            tic
                % Approximate inner products using FFT
            fftfT = fft(fdownsampT);
            fhatT = (sqrt(2*T)/Nsamp)*((-1).^(-M:M)').*[fftfT((Nsamp-M+1):Nsamp);fftfT(1:(M+1))];
                % Solve system using Tikhonov regularization
    %         [S,lambda] = myDPSS(2*M+1,1/(2*T),2*M+1);
    %         ghatTtik = S*((lambda./(lambda.^2+alpha)).*(S'*fhatT));
            B = toeplitz(sinc((0:(2*M))/T)/T);
            C = B^2;
            C(1:(2*M+2):end) = C(1:(2*M+2):end) + alpha;
            ghatTtik = C\(B*fhatT);
            Time_FourierExtensionTik(ii,jj) = toc;
            if(~isempty(lastwarn))
                flag(jj) = true;
                lastwarn('');
            end
            clear fftfT fhatT S lambda B C
        end
        
        % Compute Fourier Extension coefficients using fast pseudoinverse
        tic
%             % Approximate inner products using FFT
        fftfT = fft(fdownsampT);
        fhatT = (sqrt(2*T)/Nsamp)*((-1).^(-M:M)').*[fftfT((Nsamp-M+1):Nsamp);fftfT(1:(M+1))];
%             % Solve system using fast truncated pseudoinverse
        FPP = FastProlatePseudoinverse(2*M+1,1/(2*T),cutoff,tolerance);
        ghatTpinvFast = FPP.FPpinv(fhatT);
        Time_FourierExtensionPinvFast(ii,jj) = toc;
        if(~isempty(lastwarn))
            flag(jj) = true;
            lastwarn('');
        end
        clear fftfT FPP fhatT
        
        % Compute Fourier Extension coefficients using fast Tikhonov regularization
        tic
            % Approximate inner products using FFT
        fftfT = fft(fdownsampT);
        fhatT = (sqrt(2*T)/Nsamp)*((-1).^(-M:M)').*[fftfT((Nsamp-M+1):Nsamp);fftfT(1:(M+1))];
            % Solve system using fast Tikhonov regularization
        FPT = FastProlateTikhonov(2*M+1,1/(2*T),alpha,tolerance);
        ghatTtikFast = FPT.FPtik(fhatT);
        Time_FourierExtensionTikFast(ii,jj) = toc;
        if(~isempty(lastwarn))
            flag(jj) = true;
            lastwarn('');
        end        
        clear fftfT FPT fhatT
        
        % Only compute errors once, since algorithms are deterministic
        if(ii == 1)
            % f_M(t) over [-1,1] gridpoints
            fMsamp = (N/sqrt(2))*ifft([fhat((M+1):(2*M+1)).*((-1).^(0:M)');zeros(N-2*M-1,1);fhat(1:M).*((-1).^(-M:-1)')]);
            
            % g_M(t) over [-T,T] gridpoints
            if(M < 4096)
                gMsampTpinv = (N/sqrt(2*T))*ifft([ghatTpinv((M+1):(2*M+1)).*((-1).^(0:M)');zeros(N-2*M-1,1);ghatTpinv(1:M).*((-1).^(-M:-1)')]);            
                gMsampTtik = (N/sqrt(2*T))*ifft([ghatTtik((M+1):(2*M+1)).*((-1).^(0:M)');zeros(N-2*M-1,1);ghatTtik(1:M).*((-1).^(-M:-1)')]);
            end
            gMsampTpinvFast = (N/sqrt(2*T))*ifft([ghatTpinvFast((M+1):(2*M+1)).*((-1).^(0:M)');zeros(N-2*M-1,1);ghatTpinvFast(1:M).*((-1).^(-M:-1)')]);            
            gMsampTtikFast = (N/sqrt(2*T))*ifft([ghatTtikFast((M+1):(2*M+1)).*((-1).^(0:M)');zeros(N-2*M-1,1);ghatTtikFast(1:M).*((-1).^(-M:-1)')]);

            % Compute Relative RMS Errors
            Error_Fourier(jj) = rms(fMsamp-fsamp);
            if(M < 4096)
                Error_FourierExtensionPinv(jj) = rms(gMsampTpinv(indexT)-fsampT(indexT));
                Error_FourierExtensionTik(jj) = rms(gMsampTtik(indexT)-fsampT(indexT));
            end
            Error_FourierExtensionPinvFast(jj) = rms(gMsampTpinvFast(indexT)-fsampT(indexT));
            Error_FourierExtensionTikFast(jj) = rms(gMsampTtikFast(indexT)-fsampT(indexT));
            
            clear fMsamp gMsampTpinv gMsampTtik gMsampTpinvFast gMsampTtikFast
        end
        
        clear fhat ghatTpinv ghatTtik ghatTpinvFast ghatTtikFast
    end
    if(ii == 1)
        % Normalize RMS Errors
        Error_Fourier = Error_Fourier/rms(fsamp);
        temp = rms(fsampT(indexT));
        Error_FourierExtensionPinv = Error_FourierExtensionPinv/temp;
        Error_FourierExtensionTik = Error_FourierExtensionTik/temp;
        Error_FourierExtensionPinvFast = Error_FourierExtensionPinvFast/temp;
        Error_FourierExtensionTikFast = Error_FourierExtensionTikFast/temp; 
    end
end

% Save data
save('FastFourierExtension.mat','T','cutoff','alpha','tolerance','K','mu','sigma','a','N','vecM','numtrials','flag',...
'Error_Fourier','Error_FourierExtensionPinv','Error_FourierExtensionTik','Error_FourierExtensionPinvFast','Error_FourierExtensionTikFast',...
'Time_Fourier','Time_FourierExtensionPinv','Time_FourierExtensionTik','Time_FourierExtensionPinvFast','Time_FourierExtensionTikFast')