function FPT = FastProlateTikhonov(N,W,alpha,epsilon)
%% FastProlateTikhonov.m 
% Function implementing fast Tikhonov regularization involving B_{N,W} as 
% described in "The Fast Slepian Transform". For a regularization parameter
% alpha > 0, the matrix B_{N,W}^{(tik)} = (B_{N,W}^2+alpha*I)^{-1}B_{N,W}
% can be approximated as the prolate matrix B_{N,W} plus a low rank
% correction matrix.
% 
% Inputs:
% -N: Length of DPSSs
% -W: Half Bandwidth for DPSSs
% -alpha: Regularization parameter
% -epsilon: Relative error in approximation
%
% Outputs:
% -FPT an object containing a Fast Prolate Tikhonov function
%
% Given an Nx1 vector x, performing
% z = FPT.FPtik(x);
% results in z being an Nx1 vector such that ||z-B_{N,W}^{(tik)}x||_2 <= epsilon*||x||_2
% If x is an NxM array, then FPP.FPtik() operates columnwise.
%
% Most recent change - 08/07/2017
%
% Copyright 2017, Santhosh Karnik
%
% This file is part of Fast Slepian Transform (FST) Toolbox version 1.0.
%
%    FST is free software: you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation, either version 3 of the License, or
%    (at your option) any later version.
%
%    FST is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License
%    along with FST.  If not, see <http://www.gnu.org/licenses/>.
%
% If you use this code in an academic paper, please cite our paper:
% S. Karnik, Z. Zhu, M. B. Wakin, J. K. Romberg, and M. A. Davenport., 
% The fast Slepian transform, Appl. Comput. Harmon. Anal. (2017), 
% http://dx.doi.org/10.1016/j.acha.2017.07.005

%% Fast Slepian Projection Function
FPT.FPtik = @FPtik;

%% Compute DPSSs and store transition region DPSSs
[FPT.S,lambda,lowerindex,upperindex] = transitionDPSS(N,W,[min(alpha*(1+alpha)*epsilon,1/4);epsilon/3]);
FPT.eig_weights = lambda.*(1./(lambda.^2+alpha)-1./(1+alpha));

%% Precompuute length 2^q FFT of sinc function needed to apply B_{N,W} to a vector (where 2^q >= 2*N-1).
FPT.fftlength = 2^ceil(log2(2*N-1));
vecScaledSinc = 2*W*sinc(2*W*(0:(N-1))')/(1+alpha);
FPT.scaledsincFFT = fft([vecScaledSinc;...
                   zeros(FPT.fftlength-2*N+1,1);...
                   vecScaledSinc(end:-1:2)]);

%% Store signal length, bandwidth, tolerance, and number of DPSSs
FPT.N = N; % Signal Length
FPT.W = W; % Half-Bandwidth
FPT.alpha = alpha; % Regularization Parameter
FPT.epsilon = epsilon; % Tolerance

%% Fast Prolate Pseudoinverse
function Btikx = FPtik(x)
    if(FPT.N ~= size(x,1))
        disp('ERROR: Input has incorrect length')
    else
        % Toeplitz matrix multiply (Use FFT of length 2^q >= 2*N-1)
        Bx = ifft(bsxfun(@times,FPT.scaledsincFFT,fft(x,FPT.fftlength,1)));
        
        % Low rank update
        SDSx = (FPT.S)*bsxfun(@times,FPT.eig_weights,(FPT.S)'*x);
        
        % Add Toeplitz*vector to low rank update
        Btikx = Bx(1:FPT.N,:)+SDSx;
    end
end

end